# Copyright 2017 QuantRocket LLC - All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd
from .exceptions import InsufficientData

class Performance(object):
    """
    Class for storing performance attributes and calculating derived statistics.

    Parameters
    ----------
    returns : DataFrame, required
        a Dataframe of pct returns

    pnl : DataFrame, optional
        a DataFrame of pnl

    net_exposures : DataFrame, optional
        a Dataframe of net (hedged) exposure

    abs_exposures : DataFrame, optional
        a Dataframe of absolute exposure (ignoring hedging)

    commissions : DataFrame, optional
        a DataFrame of commissions, in the base currency

    commissions_pct : DataFrame, optional
        a DataFrame of commissions, in percentages

    slippages : DataFrame, optional
        a DataFrame of slippages, in percentages

    benchmark : Series, optional
        a Series of prices for a benchmark

    riskfree : float, optional
        the riskfree rate (default 0)

    compound_returns : bool
         True for compound/geometric returns, False for arithmetic returns (default True)

    rolling_sharpe_window : int, optional
        compute rolling Sharpe over this many periods (default 200)
    """

    def __init__(
        self,
        returns,
        pnl=None,
        net_exposures=None,
        abs_exposures=None,
        commissions=None,
        commissions_pct=None,
        slippages=None,
        benchmark=None,
        riskfree=0,
        compound_returns=True,
        rolling_sharpe_window=200
        ):

        self.returns = returns
        if len(self.returns.index) < 2:
            raise InsufficientData(
                "Moonchart needs at least 2 dates to analyze performance, "
                "but returns DataFrame has length {0}".format(len(self.returns.index)))
        self.pnl = pnl
        self.net_exposures = net_exposures
        self.abs_exposures = abs_exposures
        self.commissions = commissions
        self.commissions_pct = commissions_pct
        self.slippages = slippages
        self.benchmark = benchmark
        self.riskfree = riskfree
        self.rolling_sharpe_window = rolling_sharpe_window
        self.compound_returns = compound_returns
        self.returns_with_baseline = None
        self.cum_returns = None
        self.cum_returns_with_baseline = None
        self.sharpe = None
        self.rolling_sharpe = None
        self.cagr = None
        self.drawdowns = None
        self.max_drawdown = None
        self.cum_pnl = None
        self._performance_cache_filled = False

    @classmethod
    def from_moonshot(cls, results):
        """
        Creates a Performance instance from a moonshot backtest results DataFrame.
        """
        fields = results.index.get_level_values("Field").unique()
        kwargs = {}
        kwargs["returns"] = results.loc["Return"].astype(np.float64)
        if "NetExposure" in fields:
            kwargs["net_exposures"] = results.loc["NetExposure"].astype(np.float64)
        if "AbsExposure" in fields:
            kwargs["abs_exposures"] = results.loc["AbsExposure"].astype(np.float64)
        if "Commission" in fields:
            kwargs["commissions_pct"] = results.loc["Commission"].astype(np.float64)
        if "Slippage" in fields:
            kwargs["slippages"] = results.loc["Slippage"].astype(np.float64)
        if "Benchmark" in fields:
            kwargs["benchmark"] = results.loc["Benchmark"].astype(np.float64)

        return cls(**kwargs)

    @classmethod
    def from_pnl(cls, results):
        """
        Creates a Performance instance from a PNL results DataFrame.
        """
        fields = results.index.get_level_values("Field").unique()
        kwargs = {}
        kwargs["returns"] = results.loc["Return"].astype(np.float64)
        kwargs["pnl"] = results.loc["Pnl"].astype(np.float64)
        if "NetExposure" in fields:
            kwargs["net_exposures"] = results.loc["NetExposure"].astype(np.float64)
        if "AbsExposure" in fields:
            kwargs["abs_exposures"] = results.loc["AbsExposure"].astype(np.float64)
        if "Commission" in fields:
            kwargs["commissions_pct"] = results.loc["Commission"].astype(np.float64)
        if "CommissionAmount" in fields:
            kwargs["commissions"] = results.loc["CommissionAmount"].astype(np.float64)
        if "Benchmark" in fields:
            kwargs["benchmark"] = results.loc["Benchmark"].astype(np.float64)

        return cls(**kwargs)

    def fill_performance_cache(self):
        if self._performance_cache_filled:
            return

        self.cum_returns = self.get_cum_returns(self.returns)
        self.cum_returns_with_baseline = self.get_cum_returns(
            self.with_baseline(self.returns))
        self.cagr = self.get_cagr(self.cum_returns)
        self.sharpe = self.get_sharpe(self.returns)
        self.rolling_sharpe = self.get_rolling_sharpe(self.returns)
        self.drawdowns = self.get_drawdowns(self.cum_returns)
        self.max_drawdown = self.get_max_drawdown(self.drawdowns)
        if self.pnl is not None:
            self.cum_pnl = self.pnl.cumsum()

        self._performance_cache_filled = True

    def with_baseline(self, data):
        """
        Adds an initial period with a return of 0, as a baseline.
        """
        period_length = data.index[1] - data.index[0]
        prior_period = data.index[0] - period_length
        if isinstance(data, pd.DataFrame):
            baseline_row = pd.DataFrame(0, index=[prior_period], columns=data.columns)
        else:
            baseline_row = pd.Series(0, index=[prior_period], name=data.name)
        data_with_baseline = pd.concat((baseline_row, data))
        return data_with_baseline

    def get_sharpe(self, returns):
        """
        Returns the Sharpe ratio of the provided returns (which should be a
        DataFrame or Series).
        """
        mean = (returns - self.riskfree).mean()
        if isinstance(mean, float) and mean == 0:
            return 0
        std = (returns - self.riskfree).std()
        # Returns are assumed to represent daily returns, so annualize the Sharpe ratio
        return mean/std * np.sqrt(252)

    def get_rolling_sharpe(self, returns):
        """
        Computes rolling Sharpe ratios for the returns. Returns should be a
        DataFrame.
        """
        rolling_returns = returns.fillna(0).rolling(
            self.rolling_sharpe_window, min_periods=self.rolling_sharpe_window)
        try:
            return rolling_returns.apply(self.get_sharpe, raw=True)
        except TypeError as e:
            # handle pandas<0.23
            if "apply() got an unexpected keyword argument 'raw'" in repr(e):
                return rolling_returns.apply(self.get_sharpe)
            else:
                raise

    def get_cum_returns(self, returns, compound=None):
        """
        Computes the cumulative returns of the provided Series or DataFrame.
        """
        if compound is None:
            compound = self.compound_returns
        if compound:
            cum_returns = (1 + returns).cumprod()
        else:
            cum_returns = returns.cumsum() + 1

        cum_returns.index.name = "Date"
        return cum_returns

    def get_cagr(self, cum_returns):
        """
        Computes the CAGR of the cum_returns (a DataFrame or Series).
        """
        # For DataFrames, apply this method to each Series.
        if isinstance(cum_returns, pd.DataFrame):
            return cum_returns.apply(self.get_cagr, axis=0)

        # Ignore nulls when compting CAGR
        cum_returns = cum_returns[cum_returns.notnull()]

        if cum_returns.empty:
            return 0

        # Compute the CAGR of the Series
        min_date = cum_returns.index.min()
        max_date = cum_returns.index.max()
        years = ((max_date - min_date).days or 1)/365.0
        ending_value = cum_returns.iloc[-1]
        # Since we are computing CAGR on cumulative returns, the beginning
        # value is always 1.
        beginning_value = 1
        if self.compound_returns:
            cagr = (ending_value/beginning_value)**(1/years) - 1
        else:
            # Compound annual growth rate doesn't apply to arithmetic
            # returns, so just divide the cum_returns by the number of years
            # to get the annual return
            cagr = (ending_value/beginning_value - 1)/years

        return cagr

    def get_avg_exposure(self, exposures):
        """
        Calculates the avg exposure.
        """
        return exposures.mean()

    def get_normalized_cagr(self, cagr, exposure):
        """
        Returns the CAGR per 1x exposure, a measure of the strategy's
        efficiency.
        """
        return cagr / exposure

    def get_drawdowns(self, cum_returns):
        """
        Computes the drawdowns of the cum_returns (a Series or DataFrame).
        """
        cum_returns = cum_returns[cum_returns.notnull()]
        highwater_marks = cum_returns.expanding().max()
        drawdowns = cum_returns/highwater_marks - 1
        return drawdowns

    def get_max_drawdown(self, drawdowns):
        """
        Returns the max drawdown.
        """
        return drawdowns.min()

    def get_benchmark_returns(self):
        """
        Returns a Series of benchmark prices, if any. If more than one strategy/column has
        benchmark prices, returns the first.
        """
        if self.benchmark is None:
            return None
        have_benchmarks = self.benchmark.notnull().any(axis=0)
        have_benchmarks = have_benchmarks[have_benchmarks]
        if have_benchmarks.empty:
            return None

        col = have_benchmarks.index[0]
        if len(have_benchmarks.index) > 1:
            import warnings
            warnings.warn("Multiple benchmarks found, only using first ({0})".format(col))

        benchmark_prices = self.benchmark[col]
        benchmark_prices.name = "benchmark"
        return benchmark_prices.pct_change()

    def get_top_movers(self, returns, top_n=10):
        """
        Returns the biggest gainers and losers in the returns.
        """

        if isinstance(returns, pd.DataFrame):
            returns = returns.stack()

        returns = returns.sort_values()

        return pd.concat((returns.head(top_n), returns.tail(top_n)))

class AggregatePerformance(Performance):

    def __init__(self, performance):

        super(AggregatePerformance, self).__init__(
            performance.returns.sum(axis=1),
            riskfree=performance.riskfree,
            compound_returns=performance.compound_returns,
            rolling_sharpe_window=performance.rolling_sharpe_window,
            benchmark=performance.benchmark
        )
        if performance.pnl is not None:
            self.pnl = performance.pnl.sum(axis=1)

        if performance.commissions is not None:
            self.commissions = performance.commissions.sum(axis=1)

        if performance.commissions_pct is not None:
            self.commissions_pct = performance.commissions_pct.sum(axis=1)

        if performance.slippages is not None:
            self.slippages = performance.slippages.sum(axis=1)

        if performance.net_exposures is not None:
            self.net_exposures = performance.net_exposures.sum(axis=1)

        if performance.abs_exposures is not None:
            self.abs_exposures = performance.abs_exposures.sum(axis=1)



# Copyright 2017 QuantRocket LLC - All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
import matplotlib.pyplot as plt
from .base import BaseTearsheet

class ShortFallTearsheet(BaseTearsheet):

    def create_shortfall_tearsheet(self, live_returns, simulated_returns):
        """
        Shows drawdown and equity curves comparing the live and simulated
        returns. DataFrames should include a column for each strategy being
        analyzed.
        """
        raise NotImplementedError()

        # TODO: be more specific
        self.suptitle = "Shortfall Analysis"

        strategies = live_returns.columns.union(simulated_returns.columns)
        strategy_count = len(strategies)

        rows, cols = self._get_plot_dimensions(strategy_count)

        for i, strategy in enumerate(strategies):
            if strategy in live_returns.columns:
                strategy_live_returns = live_returns[strategy]
            else:
                strategy_live_returns = pd.Series(0, index=live_returns.index)
            strategy_live_returns.name = "live"
            strategy_simulated_returns = simulated_returns[strategy]
            strategy_simulated_returns.name = "simulated"
            strategy_returns = pd.concat((strategy_live_returns, strategy_simulated_returns), axis=1).fillna(0)
            cum_returns = self.cum_returns(self._with_baseline(strategy_returns))
            drawdowns = self.drawdowns(cum_returns)
            shortfall = cum_returns.live - cum_returns.simulated

            fig = plt.figure("Cumulative Returns", figsize=self.window_size)
            fig.suptitle(self.suptitle, **self.suptitle_kwargs)
            axis = fig.add_subplot(rows, cols, i+1)
            cum_returns.plot(ax=axis, title=strategy)

            fig = plt.figure("Drawdowns", figsize=self.window_size)
            fig.suptitle(self.suptitle, **self.suptitle_kwargs)
            axis = fig.add_subplot(rows, cols, i+1)
            drawdowns.plot(ax=axis, title=strategy)

            fig = plt.figure("Shortfall", figsize=self.window_size)
            fig.suptitle(self.suptitle, **self.suptitle_kwargs)
            axis = fig.add_subplot(rows, cols, i+1)
            shortfall.plot(ax=axis, title=strategy, kind="area", stacked=False)

        self._save_or_show()


# Copyright 2017 QuantRocket LLC - All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from matplotlib import cycler
from matplotlib.backends.backend_pdf import PdfPages

# Set seaborn default style
sns.set()

class BaseTearsheet(object):
    """
    Base class for tear sheets.
    """
    DEFAULT_TITLE = "Performance tear sheet"

    def __init__(self, pdf_filename=None, window_size=None, max_cols_for_details=25):
        self.window_size = window_size or (12.0, 7.5) # width, height in inches
        plt.rc("legend", fontsize="small")
        plt.rc("axes", axisbelow=True)
        plt.rc("figure", autolayout=True)
        plt.rc("xtick", labelsize="small")
        plt.rc("ytick", labelsize="small")
        if pdf_filename:
            self.pdf = PdfPages(pdf_filename, keep_empty=True)
        else:
            self.pdf = None

        self.suptitle = self.DEFAULT_TITLE
        self.suptitle_kwargs = {
            "bbox": dict(facecolor="#EAEAF2", edgecolor='white', alpha=0.5)}
        self.max_cols_for_details = max_cols_for_details

    def _save_or_show(self):
        """
        Saves the fig to the multi-page PDF, or shows it.
        """
        if self.pdf:
            for fignum in plt.get_fignums():
                self.pdf.savefig(fignum)
            plt.close("all")
            self.pdf.close()
        else:
            plt.show()

    def _get_plot_dimensions(self, plot_count):
        """
        Returns a tuple of rows, cols needed to accomodate the plot_count.
        """
        rows = math.ceil(math.sqrt(plot_count))
        cols = math.ceil(plot_count/rows)
        return rows, cols

    @property
    def _tight_layout_clear_suptitle(self):
        # leave room at top for suptitle
        return dict(rect=[0,0,1,.9])

    def _clear_legend(self, plot, legend_title=None):
        """
        Anchors the legend to the outside right of the plot area so you can
        see the plot.
        """
        plot.figure.set_tight_layout({"pad":10, "h_pad":1, "w_pad":1})
        plot.legend(
            loc='center left', bbox_to_anchor=(1, 0.5), fontsize="x-small", title=legend_title)

    def _create_performance_plots(self, performance, subplot, extra_label):
        """
        Creates agg/details plots for cumulative returns, drawdowns, rolling
        Sharpe, and possibly pnl.
        """
        if subplot == 111:
            tight_layout = self._tight_layout_clear_suptitle
        else:
            tight_layout = None

        fig = plt.figure("Cumulative Returns", figsize=self.window_size, tight_layout=tight_layout)
        fig.suptitle(self.suptitle, **self.suptitle_kwargs)
        axis = fig.add_subplot(subplot)
        max_return = performance.cum_returns.max(axis=0)
        if isinstance(max_return, pd.Series):
            max_return = max_return.max(axis=0)
        # If the price more than doubled, use a log scale
        if max_return >= 2:
            axis.set_yscale("log", basey=2)

        include_commissions = (
            performance.commissions_pct is not None
            # if all commissions are null/0, don't show them
            and (performance.commissions_pct.fillna(0) != 0).any())

        include_slippage = (
            performance.slippages is not None
            # if all slippages are null/0, don't show them
            and (performance.slippages.fillna(0) != 0).any())

        if (
            # a 212 subplot means a detailed plot, which isn't compatible with
            # showing commissions and slippage
            subplot != 212 and (include_commissions or include_slippage)):

            if include_commissions:
                commissions_pct = performance.with_baseline(performance.commissions_pct)
                cum_commissions_pct = performance.get_cum_returns(commissions_pct)
                cum_commissions_pct.name = "commissions"

            if include_slippage:
                slippages = performance.with_baseline(performance.slippages)
                cum_slippages = performance.get_cum_returns(slippages)
                cum_slippages.name = "slippage"

            performance.cum_returns_with_baseline.name = "returns"

            cum_gross_returns = performance.cum_returns_with_baseline

            if include_commissions:
                cum_gross_returns = cum_gross_returns.multiply(cum_commissions_pct)

            if include_slippage:
                cum_gross_returns = cum_gross_returns.multiply(cum_slippages)

            cum_gross_returns.name = "gross returns"
            breakdown_parts = [performance.cum_returns_with_baseline, cum_gross_returns]

            if include_commissions:
                breakdown_parts.append(cum_commissions_pct)

            if include_slippage:
                breakdown_parts.append(cum_slippages)

            returns_breakdown = pd.concat(breakdown_parts, axis=1)
            plot = returns_breakdown.plot(ax=axis, title="Cumulative Returns {0}".format(extra_label))
            if isinstance(returns_breakdown, pd.DataFrame):
                self._clear_legend(plot)
        else:
            plot = performance.cum_returns_with_baseline.plot(ax=axis, title="Cumulative Returns {0}".format(extra_label))
            if isinstance(performance.cum_returns_with_baseline, pd.DataFrame):
                self._clear_legend(plot)

        fig = plt.figure("Drawdowns", figsize=self.window_size, tight_layout=tight_layout)
        fig.suptitle(self.suptitle, **self.suptitle_kwargs)
        axis = fig.add_subplot(subplot)
        plot = performance.drawdowns.plot(ax=axis, title="Drawdowns {0}".format(extra_label))
        if isinstance(performance.drawdowns, pd.DataFrame):
            self._clear_legend(plot)

        if performance.cum_pnl is not None:
            # Cumulative pnl will generally be very similar to cumulative
            # return, but they might be different if cumulative return is
            # calcuated based on an account balance that changes due to
            # causes other than the strategies being charted (e.g. other
            # strategies, contributions/withdrawals, etc.)
            fig = plt.figure("Cumulative PNL", figsize=self.window_size,
                             tight_layout=self._tight_layout_clear_suptitle)
            fig.suptitle(self.suptitle, **self.suptitle_kwargs)
            axis = fig.add_subplot(subplot)
            if (
                performance.commissions is not None
                # a 212 subplot means a detailed plot, which isn't compatible with
                # showing commissions
                and subplot != 212
                # if all commissions are null/0, don't show them
                and (performance.commissions.fillna(0) != 0).any()):
                cum_commissions = performance.commissions.cumsum()
                cum_commissions.name = "commissions"
                cum_pnl = performance.pnl.cumsum()
                cum_pnl.name = "pnl"
                cum_gross_pnl = cum_pnl + cum_commissions.abs()
                cum_gross_pnl.name = "gross pnl"
                pnl_breakdown = pd.concat((cum_pnl, cum_gross_pnl, cum_commissions), axis=1)
                plot = pnl_breakdown.plot(ax=axis, title="Cumulative PNL {0}".format(extra_label))
                if isinstance(returns_breakdown, pd.DataFrame):
                    self._clear_legend(plot)
            else:
                plot = performance.cum_pnl.plot(ax=axis, title="Cumulative PNL {0}".format(extra_label))
                if isinstance(performance.cum_pnl, pd.DataFrame):
                    self._clear_legend(plot)

        if len(performance.rolling_sharpe.index) > performance.rolling_sharpe_window:
            fig = plt.figure("Rolling Sharpe", figsize=self.window_size, tight_layout=tight_layout)
            fig.suptitle(self.suptitle, **self.suptitle_kwargs)
            axis = fig.add_subplot(subplot)
            plot = performance.rolling_sharpe.plot(ax=axis, title="Rolling Sharpe ({0}-day) {1}".format(
                performance.rolling_sharpe_window, extra_label))
            if isinstance(performance.rolling_sharpe, pd.DataFrame):
                self._clear_legend(plot)

        benchmark_returns = performance.get_benchmark_returns()
        if benchmark_returns is not None:
            fig = plt.figure("Cumulative Returns vs Benchmark", figsize=self.window_size, tight_layout=tight_layout)
            fig.suptitle(self.suptitle, **self.suptitle_kwargs)
            axis = fig.add_subplot(subplot)
            max_return = performance.cum_returns.max(axis=0)
            if isinstance(max_return, pd.Series):
                max_return = max_return.max(axis=0)
            # If the price more than doubled, use a log scale
            if max_return >= 2:
                axis.set_yscale("log", basey=2)

            if isinstance(performance.cum_returns_with_baseline, pd.Series):
                performance.cum_returns_with_baseline.name = "strategy"
            benchmark_cum_returns = performance.get_cum_returns(performance.with_baseline(benchmark_returns))
            vs_benchmark = pd.concat((performance.cum_returns_with_baseline, benchmark_cum_returns), axis=1)
            plot = vs_benchmark.plot(ax=axis, title="Cumulative Returns vs Benchmark {0}".format(extra_label))
            if isinstance(vs_benchmark, pd.DataFrame):
                self._clear_legend(plot)



# Copyright 2017 QuantRocket LLC - All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd
import seaborn as sns
from collections import OrderedDict
import matplotlib.pyplot as plt
import warnings
from .perf import Performance, AggregatePerformance
from .base import BaseTearsheet

class Tearsheet(BaseTearsheet):
    """
    Generates a tear sheet of performance stats and graphs.
    """

    def _set_title_from_performance(self, performance):
        """
        Sets a title like "<start date> - <end date>: <securities/strategies/columns>"
        """
        min_date = performance.returns.index.min().date().isoformat()
        max_date = performance.returns.index.max().date().isoformat()
        cols = list(performance.returns.columns)
        cols = ", ".join([str(col) for col in cols])
        if len(cols) > 70:
            cols = cols[:70] + "..."
        self.suptitle = "{0} - {1}: {2}".format(
            min_date, max_date, cols)

    def from_moonshot(self, results, **kwargs):
        """
        Creates a full tear sheet from a moonshot backtest results DataFrame.

        Parameters
        ----------
        results : DataFrame
            multiindex (Field, Date) DataFrame of backtest results

        Returns
        -------
        None
        """
        performance = Performance.from_moonshot(results)
        return self.create_full_tearsheet(performance, **kwargs)

    def from_moonshot_csv(self, filepath_or_buffer, **kwargs):
        """
        Creates a full tear sheet from a moonshot backtest results CSV.

        Parameters
        ----------
        filepath_or_buffer : str or file-like object
            filepath or file-like object of the CSV

        Returns
        -------
        None
        """
        results = pd.read_csv(filepath_or_buffer,
                              parse_dates=["Date"],
                              index_col=["Field","Date"])
        return self.from_moonshot(results, **kwargs)

    def from_pnl(self, results, **kwargs):
        """
        Creates a full tear sheet from a pnl DataFrame.

        Parameters
        ----------
        results : DataFrame
            multiindex (Field, Date) DataFrame of performance results

        Returns
        -------
        None
        """
        performance = Performance.from_pnl(results)
        return self.create_full_tearsheet(performance, **kwargs)

    def from_pnl_csv(self, filepath_or_buffer, **kwargs):
        """
        Creates a full tear sheet from a pnl CSV.

        Parameters
        ----------
        filepath_or_buffer : str or file-like object
            filepath or file-like object of the CSV

        Returns
        -------
        None
        """
        results = pd.read_csv(filepath_or_buffer,
                              parse_dates=["Date"],
                              index_col=["Field","Date"])
        return self.from_pnl(results, **kwargs)

    def create_full_tearsheet(
        self,
        performance,
        include_exposures_tearsheet=True,
        include_annual_breakdown_tearsheet=True,
        montecarlo_n=None,
        montecarlo_preaggregate=True,
        title=None
        ):
        """
        Create a full tear sheet of performance results and market exposure.

        Parameters
        ----------
        performance : instance
            Performance instance

        include_exposures : bool
            whether to include a tear sheet of market exposure

        include_annual_breakdown_tearsheet : bool
            whether to include an annual breakdown of Sharpe and CAGR

        montecarlo_n : int
            how many Montecarlo simulations to run on the returns, if any

        montecarlo_preaggregate : bool
            whether Montecarlo simulator should preaggregate returns;
            ignored unless montecarlo_n is nonzero

        Returns
        -------
        None
        """
        if title:
            self.suptitle = title
        else:
            self._set_title_from_performance(performance)

        agg_performance = AggregatePerformance(performance)

        num_cols = len(performance.returns.columns)
        if num_cols > self.max_cols_for_details:
            warnings.warn("Suppressing details because there are more than {0} columns".format(
                self.max_cols_for_details))

        self.create_performance_tearsheet(performance, agg_performance)

        if include_annual_breakdown_tearsheet:
            self.create_annual_breakdown_tearsheet(performance, agg_performance)

        if include_exposures_tearsheet and any([exposures is not None for exposures in (
            performance.net_exposures, performance.abs_exposures)]):
            self.create_exposures_tearsheet(performance, agg_performance)

        if montecarlo_n:
            self.montecarlo_simulate(
                performance, n=montecarlo_n, preaggregate=montecarlo_preaggregate)

        self._save_or_show()

    def create_performance_tearsheet(self, performance, agg_performance):
        """
        Creates a performance tearsheet.
        """
        agg_performance.fill_performance_cache()

        num_cols = len(performance.returns.columns)
        show_details = num_cols > 1 and num_cols <= self.max_cols_for_details

        if show_details:
            performance.fill_performance_cache()

        self._create_agg_performance_textbox(agg_performance)

        self._create_performance_plots(
            agg_performance,
            subplot=211 if show_details else 111,
            extra_label="(Aggregate)" if show_details else "")

        if show_details:
            self._create_performance_plots(performance, subplot=212, extra_label="(Details)")
            self._create_detailed_performance_bar_charts(performance, extra_label="(Details)")

    def _create_detailed_performance_bar_charts(self, performance, extra_label):
        if performance.pnl is not None:
            fig = plt.figure("PNL {0}".format(extra_label), figsize=self.window_size,
                             tight_layout=self._tight_layout_clear_suptitle)
            fig.suptitle(self.suptitle, **self.suptitle_kwargs)
            axis = fig.add_subplot(111)
            pnl = performance.pnl.sum().sort_values(inplace=False)
            if performance.commissions is not None:
                pnl.name = "pnl"
                commissions = performance.commissions.sum()
                commissions.name = "commissions"
                gross_pnl = pnl + commissions
                gross_pnl.name = "gross pnl"
                pnl = pd.concat((pnl, gross_pnl, commissions), axis=1)
            pnl.plot(
                ax=axis, kind="bar", title="PNL {0}".format(extra_label))

        fig = plt.figure("CAGR {0}".format(extra_label), figsize=self.window_size,
                         tight_layout=self._tight_layout_clear_suptitle)
        fig.suptitle(self.suptitle, **self.suptitle_kwargs)
        axis = fig.add_subplot(111)
        performance.cagr.sort_values(inplace=False).plot(
            ax=axis, kind="bar", title="CAGR {0}".format(extra_label))

        fig = plt.figure("Sharpe {0}".format(extra_label), figsize=self.window_size,
                         tight_layout=self._tight_layout_clear_suptitle)
        fig.suptitle(self.suptitle, **self.suptitle_kwargs)
        axis = fig.add_subplot(111)
        performance.sharpe.sort_values(inplace=False).plot(
            ax=axis, kind="bar", title="Sharpe {0}".format(extra_label))

        fig = plt.figure("Max Drawdown {0}".format(extra_label), figsize=self.window_size,
                         tight_layout=self._tight_layout_clear_suptitle)
        fig.suptitle(self.suptitle, **self.suptitle_kwargs)
        axis = fig.add_subplot(111)
        performance.max_drawdown.sort_values(inplace=False).plot(
            ax=axis, kind="bar", title="Max drawdown {0}".format(extra_label))

    def _create_agg_performance_textbox(self, agg_performance):
        agg_stats = OrderedDict()
        agg_stats_text = ""

        if agg_performance.pnl is not None:
            agg_stats["PNL"] = round(agg_performance.pnl.sum(), 4)
        if agg_performance.commissions is not None:
            agg_stats["Commissions"] = round(agg_performance.commissions.sum(), 4)

        agg_stats["CAGR"] = agg_performance.cagr
        agg_stats["Sharpe"] = agg_performance.sharpe
        agg_stats["Max Drawdown"] = agg_performance.max_drawdown

        agg_stats_text = self._get_agg_stats_text(agg_stats)
        fig = plt.figure("Aggregate Performance", figsize=self.window_size)
        fig.suptitle(self.suptitle, **self.suptitle_kwargs)
        self.plot_textbox(fig, agg_stats_text)

    def plot_textbox(self, fig, text):
        with sns.axes_style("white", {'axes.linewidth': 0}):
            axis = fig.add_subplot(111)
            axis.get_xaxis().set_visible(False)
            axis.get_yaxis().set_visible(False)
            axis.text(0.1, 0.4, text,
                 family="monospace",
                 fontsize="xx-large"
                 )
    def create_exposures_tearsheet(self, performance, agg_performance):
        """
        Create a tearsheet of market exposure.
        """
        agg_performance.fill_performance_cache()

        num_cols = len(performance.returns.columns)
        show_details = num_cols > 1 and num_cols <= self.max_cols_for_details
        if show_details:
            performance.fill_performance_cache()

        self._create_agg_exposures_textbox(agg_performance)

        self._create_exposures_plots(
            agg_performance,
            subplot=211 if show_details else 111,
            extra_label="(Aggregate)" if show_details else "")

        if show_details:
            self._create_exposures_plots(performance, subplot=212, extra_label="(Details)")
            self._create_detailed_exposures_bar_charts(performance, extra_label="(Details)")

    def _create_agg_exposures_textbox(self, agg_performance):

        agg_stats = OrderedDict()
        agg_stats_text = ""

        if agg_performance.net_exposures is not None:
            avg_net_exposures = agg_performance.get_avg_exposure(agg_performance.net_exposures)
            agg_stats["Avg Net Exposure"] = round(avg_net_exposures, 3)

        if agg_performance.abs_exposures is not None:
            avg_abs_exposures = agg_performance.get_avg_exposure(agg_performance.abs_exposures)
            norm_cagr = agg_performance.get_normalized_cagr(agg_performance.cagr, avg_abs_exposures)
            agg_stats["Avg Absolute Exposure"] = round(avg_abs_exposures, 3)
            agg_stats["Normalized CAGR (CAGR/Exposure)"] = round(norm_cagr, 3)

        agg_stats_text = self._get_agg_stats_text(agg_stats, title="Aggregate Exposure")
        fig = plt.figure("Aggregate Exposure", figsize=self.window_size)
        fig.suptitle(self.suptitle, **self.suptitle_kwargs)
        self.plot_textbox(fig, agg_stats_text)

    def _create_exposures_plots(self, performance, subplot, extra_label):
        if subplot == 111:
            tight_layout = self._tight_layout_clear_suptitle
        else:
            tight_layout = None

        if performance.net_exposures is not None:
            fig = plt.figure("Net Exposures", figsize=self.window_size, tight_layout=tight_layout)
            fig.suptitle(self.suptitle, **self.suptitle_kwargs)
            axis = fig.add_subplot(subplot)
            plot = performance.net_exposures.plot(ax=axis, title="Net Exposures {0}".format(extra_label))
            if isinstance(performance.net_exposures, pd.DataFrame):
                self._clear_legend(plot)

        if performance.abs_exposures is not None:
            fig = plt.figure("Absolute Exposures", figsize=self.window_size, tight_layout=tight_layout)
            fig.suptitle(self.suptitle, **self.suptitle_kwargs)
            axis = fig.add_subplot(subplot)
            plot = performance.abs_exposures.plot(ax=axis, title="Absolute Exposures {0}".format(extra_label))
            if isinstance(performance.abs_exposures, pd.DataFrame):
                self._clear_legend(plot)

    def _create_detailed_exposures_bar_charts(self, performance, extra_label):

        if performance.abs_exposures is not None:
            fig = plt.figure("Avg Absolute Exposure {0}".format(extra_label), figsize=self.window_size,
                             tight_layout=self._tight_layout_clear_suptitle)
            fig.suptitle(self.suptitle, **self.suptitle_kwargs)
            avg_abs_exposures = performance.get_avg_exposure(performance.abs_exposures)
            axis = fig.add_subplot(111)
            avg_abs_exposures.sort_values(inplace=False).plot(
                ax=axis, kind="bar", title="Avg Absolute Exposure {0}".format(extra_label))

        if performance.net_exposures is not None:
            fig = plt.figure("Avg Net Exposure {0}".format(extra_label), figsize=self.window_size,
                             tight_layout=self._tight_layout_clear_suptitle)
            fig.suptitle(self.suptitle, **self.suptitle_kwargs)
            avg_net_exposures = performance.get_avg_exposure(performance.net_exposures)
            axis = fig.add_subplot(111)
            avg_net_exposures.sort_values(inplace=False).plot(
            ax=axis, kind="bar", title="Avg Net Exposure {0}".format(extra_label))

        if performance.abs_exposures is not None:
            norm_cagrs = performance.get_normalized_cagr(performance.cagr, avg_abs_exposures)
            fig = plt.figure("Normalized CAGR (CAGR/Exposure) {0}".format(extra_label), figsize=self.window_size,
                             tight_layout=self._tight_layout_clear_suptitle)
            fig.suptitle(self.suptitle, **self.suptitle_kwargs)
            axis = fig.add_subplot(111)
            norm_cagrs.sort_values(inplace=False).plot(
                ax=axis, kind="bar", title="Normalized CAGR (CAGR/Exposure) {0}".format(extra_label))

    def create_annual_breakdown_tearsheet(self, performance, agg_performance):
        """
        Creates agg/detailed bar charts showing CAGR and Sharpe by year.
        """
        agg_performance.fill_performance_cache()

        num_cols = len(performance.returns.columns)
        show_details = num_cols > 1 and num_cols <= self.max_cols_for_details
        if show_details:
            performance.fill_performance_cache()

        self._create_annual_breakdown_plots(
            agg_performance,
            subplot=211 if show_details else 111,
            extra_label="(Aggregate)" if show_details else "")

        if show_details:
            self._create_annual_breakdown_plots(performance, subplot=212, extra_label="(Details)")

    def _create_annual_breakdown_plots(self, performance, subplot, extra_label):
        if subplot == 111:
            tight_layout = self._tight_layout_clear_suptitle
        else:
            tight_layout = None

        grouped_returns = performance.returns.groupby(performance.returns.index.year)
        cagrs_by_year = grouped_returns.apply(lambda x: performance.get_cagr(
            performance.get_cum_returns(x)))
        sharpes_by_year = grouped_returns.apply(performance.get_sharpe)

        fig = plt.figure("CAGR by Year", figsize=self.window_size, tight_layout=tight_layout)
        fig.suptitle(self.suptitle, **self.suptitle_kwargs)
        axis = fig.add_subplot(subplot)
        plot = cagrs_by_year.plot(ax=axis, kind="bar", title="CAGR by Year {0}".format(extra_label))
        if isinstance(cagrs_by_year, pd.DataFrame):
            self._clear_legend(plot)

        fig = plt.figure("Sharpe by Year", figsize=self.window_size, tight_layout=tight_layout)
        fig.suptitle(self.suptitle, **self.suptitle_kwargs)
        axis = fig.add_subplot(subplot)
        plot = sharpes_by_year.plot(ax=axis, kind="bar", title="Sharpe by Year {0}".format(extra_label))
        if isinstance(sharpes_by_year, pd.DataFrame):
            self._clear_legend(plot)

    def _get_agg_stats_text(self, agg_stats, title="Aggregate Performance"):
        """
        From a dict of aggregate stats, formats a text block.
        """
        # Create pd.Series from agg_stats to nice repr
        agg_stats = pd.Series(agg_stats)
        # Split lines
        lines = repr(agg_stats).split("\n")
        width = len(lines[0])
        # Strip last line (dtype)
        agg_stats_text = "\n".join(lines[:-1])
        agg_stats_text = "{0}\n{1}\n{2}".format(title, "="*width, agg_stats_text)
        return agg_stats_text

    def montecarlo_simulate(self, performance, n=5, preaggregate=True):
        """
        Runs a Montecarlo simulation by shuffling the dataframe of returns n
        number of times and graphing the cum_returns and drawdowns overlaid
        by the original returns. If preaggregate is True, aggregates the
        returns before the simulation, otherwise after the simulation.
        Preaggregation only randomizes by day (assuming each row is a day),
        while not preaggregating randomizes each value.
        """

        all_simulations = []

        returns = performance.returns

        if preaggregate:
            returns = returns.sum(axis=1)

        for i in range(n):
            if preaggregate:
                sim_returns = pd.Series(np.random.permutation(returns), index=returns.index)
            else:
                sim_returns = returns.apply(np.random.permutation).sum(axis=1)
            all_simulations.append(sim_returns)

        sim_returns = pd.concat(all_simulations, axis=1)

        if not preaggregate:
            returns = returns.sum(axis=1)

        cum_sim_returns = performance.get_cum_returns(performance.with_baseline(sim_returns))
        cum_returns = performance.get_cum_returns(performance.with_baseline(returns))
        fig = plt.figure("Montecarlo Simulation", figsize=self.window_size,
                         tight_layout=self._tight_layout_clear_suptitle)
        fig.suptitle(self.suptitle, **self.suptitle_kwargs)
        axis = fig.add_subplot(211)
        cum_sim_returns.plot(ax=axis, title="Montecarlo Cumulative Returns (n={0})".format(n), legend=False)
        cum_returns.plot(ax=axis, linewidth=4, color="black")
        axis = fig.add_subplot(212)
        performance.get_drawdowns(cum_sim_returns).plot(ax=axis, title="Montecarlo Drawdowns (n={0})".format(n), legend=False)
        performance.get_drawdowns(cum_returns).plot(ax=axis, linewidth=4, color="black")



# Copyright 2017 QuantRocket LLC - All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
import seaborn as sns
from collections import OrderedDict
import matplotlib
import matplotlib.pyplot as plt
from .perf import Performance
from .base import BaseTearsheet

class ParamscanTearsheet(BaseTearsheet):

    def from_moonshot(self, results, **kwargs):
        """
        Creates a param scan tear sheet from a moonshot param scan results
        DataFrame.

        Parameters
        ----------
        results : DataFrame
            DataFrame of paramscan results with columns
            (Field, StrategyOrDate, param1[, param2], Value)

        Returns
        -------
        None
        """
        idx_cols=list(results.columns)
        idx_cols.remove("Value")
        results = results.set_index(idx_cols)
        results = results["Value"] # 1-col DataFrame to Series
        idx_cols.remove("Field")
        idx_cols.remove("StrategyOrDate")
        params = idx_cols

        # preserve param order (pandas unstack() will lexsort)
        first_field = results.index.get_level_values("Field")[0]
        first_strategy = results.index.get_level_values("StrategyOrDate")[0]
        desired_cols = results.loc[first_field].loc[first_strategy].index

        results = results.unstack(level=params)
        results =results.reindex(columns=desired_cols)

        return self.create_full_tearsheet(results, **kwargs)

    def from_moonshot_csv(self, filepath_or_buffer, **kwargs):
        """
        Creates a full tear sheet from a moonshot paramscan results CSV.

        Parameters
        ----------
        filepath_or_buffer : str or file-like object
            filepath or file-like object of the CSV

        Returns
        -------
        None
        """
        results = pd.read_csv(filepath_or_buffer)
        return self.from_moonshot(results, **kwargs)

    def create_full_tearsheet(self, results, heatmap_2d=True):
        """
        Create a full tear sheet of param scan results.

        Parameters
        ----------
        results : DataFrame
            multi-index (Field, StrategyOrDate) DataFrame of param scan results,
            with param vals as (possibly multi-level) columns

        heatmap_2d : bool
            use heat maps for 2 paramscans; if False, use bar charts

        Returns
        -------
        None
        """
        # Set suptitle
        returns = results.loc["AggReturn"]
        returns.index = pd.to_datetime(returns.index)
        returns.index.name = "Date"
        min_date = returns.index.min().date().isoformat()
        max_date = returns.index.max().date().isoformat()

        if results.columns.nlevels == 2:
            params_title = " / ".join(results.columns.names)
        else:
            params_title = results.columns.name

        self.suptitle = "{0} ({1} - {2})".format(
            params_title, min_date, max_date)

        # Plot 1d bar charts or 2d heat maps
        if results.columns.nlevels == 2 and heatmap_2d:
            self._create_2d_heatmaps(results)
        else:
            self._create_1d_bar_charts(results)

        # Plot performance plots
        performance = Performance(returns)
        performance.fill_performance_cache()
        self._create_performance_plots(performance, subplot=111, extra_label=" (Aggregate)")

        self._save_or_show()

    def _create_1d_bar_charts(self, results):
        """
        Creates bar charts for 1d param scans.
        """
        fields = (
            ("Cagr", "CAGR"),
            ("Sharpe", "Sharpe"),
            ("MaxDrawdown", "Max Drawdown"),
            ("AbsExposure", "Absolute Exposure"),
            ("NormalizedCagr", "Normalized CAGR (CAGR/Exposure)")
        )
        fields = OrderedDict(fields)

        rows, cols = self._get_plot_dimensions(len(fields))
        # dynamically adjust window height based on number of plots
        width = max((self.window_size[0], cols*5+2))
        height = max((self.window_size[1], rows*2+3))
        fig = plt.figure("Parameter Scan Results", figsize=(width, height))
        fig.suptitle(self.suptitle, **self.suptitle_kwargs)

        for i, field in enumerate(list(fields.keys())):
            field_results = results.loc[field]
            field_results.index.name = "Strategy"
            field_results = field_results.T
            field_results.index = field_results.index.astype(str).str.wrap(10)

            axis = fig.add_subplot(rows, cols, i + 1)
            plot = field_results.plot(ax=axis, kind="bar", title=fields[field],
                                      fontsize="small")
            if isinstance(field_results, pd.DataFrame):
                self._clear_legend(plot, legend_title="Strategy")

            # Remove legend on all but the upper right subplot, to
            # clean up appearance
            is_upper_right =  i+1 == cols
            if not is_upper_right:
                plot.legend_.remove()

            # Hide x-axis label except on last row to save space
            is_last_row = (i+1) > (rows-1) * cols
            if not is_last_row:
                x_axis = axis.axes.get_xaxis()
                x_label = x_axis.get_label()
                x_label.set_visible(False)

    def _create_2d_heatmaps(self, results):
        """
        Creates heat maps for 2d param scans. There is one figure per field,
        with subplots for each strategy.
        """
        fields = (
            ("Cagr", "CAGR"),
            ("Sharpe", "Sharpe"),
            ("MaxDrawdown", "Max DD"),
            ("AbsExposure", "Abs Exposure"),
            ("NormalizedCagr", "Normalized CAGR")
        )
        fields = OrderedDict(fields)

        fig = None

        for i, (field, label) in enumerate(fields.items()):
            field_results = results.loc[field]
            field_results.index.name = "Strategy"
            field_results = field_results.T
            strategies = field_results.columns
            num_strategies = len(strategies)
            num_fields = len(fields)
            if not fig:
                rows, cols = self._get_plot_dimensions(num_strategies*num_fields)
                # dynamically adjust window height based on number of plots
                width = max((self.window_size[0], cols*5+2))
                height = max((self.window_size[1], rows*2+3))
                fig = plt.figure("Parameter Scan Heat Maps", figsize=(width, height), tight_layout=self._tight_layout_clear_suptitle)
                fig.suptitle(self.suptitle, **self.suptitle_kwargs)

            for ii, strategy in enumerate(strategies):
                strategy_results = field_results[strategy]

                title = "{0} ({1})".format(label, strategy)
                axis = fig.add_subplot(rows, cols, i*num_strategies + ii+ 1, title=title)
                strategy_results = strategy_results.unstack()
                strategy_results.index = strategy_results.index.astype(str).str.wrap(10)
                sns.heatmap(strategy_results, annot=True,
                    annot_kws={"size": 9},
                    center=0.0,
                    cbar=False,
                    ax=axis,
                    cmap=matplotlib.cm.RdYlGn)
